/*
Operservice 2 - opercommands.c
Loadable module
(C) Michael Meier & Henrik Mhe 2000-2002 - released under GPL
-----------------------------------------------------------------------------
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
------------------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include "globals.h"

/*
 First, the name of this mod.
 Second, the amount of items to show before truncating
*/

#define MODNAM      "opercommands"
#define SHOWTRUNC     500

void doobroadcastcmd(long unum, char *tail) {
  int res;          // Int for results
  long msgcount;       // Int to count users the text was send to
  char text[TMPSSIZE];    // Char for the text (might be too long)
  char cmdname[TMPSSIZE];   // Char for the command name (^^)
  char tmpstr[TMPSSIZE];    // Char for several purposes
  if (!checkauthlevel(unum,900)) { return; }
  // Simple error checking
  res = sscanf(tail, "%s %[^\n]", cmdname, text);
  if (res < 1) {
    putlog("!!! No commandname in opercommands.c !!!");
    return;
  } else if (res == 1) {
    sprintf(tmpstr, "Syntax: %s text", cmdname);
    msgtouser(unum, tmpstr);
    return;
  }
  // Preparing
  sprintf(tmpstr, "(oBroadcast) %s", text);
  msgcount = 0;
  tmpstr[350]='\0';  // Truncate to 350 chars
  msgcount=noticeallircops(tmpstr);
  // Message sent successfully
  sprintf(tmpstr, "Your message was sent to %ld operators.", msgcount);
  msgtouser(unum, tmpstr);
}

void dombroadcastcmd(long unum, char *tail) {
  userdata *udata;      // Struct for User Data as of User DB
  int res;          // Int for results
  int msgcount;       // Int to count users the text was send to
  int hcount;         // Int for a "for" loop
  char text[TMPSSIZE];    // Char for the text (might be too long)
  char cmdname[TMPSSIZE];   // Char for the command name (^^)
  char tmpstr[TMPSSIZE];    // Char for several purposes
  char tmpstr2[TMPSSIZE];   // Char for several purposes
  char mask[TMPSSIZE];    // Char for the mask we wanna broadcast to
  char target[10];      // String'y'fied numeric
  // Checking whether the user does have the right
  // auth level to be able to use this command.
  if (!checkauthlevel(unum,900)) { return; }
  // Simple error checking
  res = sscanf(tail, "%s %s %[^\n]", cmdname, mask, text);
  if (res < 1) {
    putlog("!!! No commandname in opercommands.c !!!");
    return;
  } else if (res < 3) {
    sprintf(tmpstr, "Syntax: %s mask text", cmdname);
    msgtouser(unum, tmpstr);
    return;
  }
  // Preparing
  sprintf(tmpstr, "(mBroadcast) %s", text);
  msgcount = 0;
  // Searching for users with the right mask
  for (hcount = 0; hcount < SIZEOFUL; hcount++) {
    udata = uls[hcount];
    while (udata != NULL) {
      // Check whether user is or is not matched by the mask
      // Checking for deaf-mode to prevent O from getting into
      // a message loop when PrivMSGing another service.

      sprintf(tmpstr2, "%s!%s@%s", udata->nick, udata->ident, udata->host);
      if ((match2strings(mask, tmpstr2)) && (ischarinstr('d', udata->umode) != 1)) {
        /* Two ideas, one is using the build in function, the other one
        is generally noticeing the message, second one seems to be better.
        First:
        msgtouser(udata->numeric, tmpstr);
        Second:
        */
        longtotoken(udata->numeric, target, 5);
        fprintf(sockout, "%sAAB O %s :%s\r\n", servernumeric, target, tmpstr);
          fflush(sockout);
        msgcount++;
      }
      udata = (void *)udata->next;
    }
  }
  // Message sent successfully
  sprintf(tmpstr, "Your message was sent to %i users.", msgcount);
  msgtouser(unum, tmpstr);
}


void dokillcmd(long unum, char *tail) {
  // Define :]
  userdata *udata;      // Struct for User Data as of User DB
  int res;          // Int for results
  long numeric;       // Long for the user's numeric
  char reason[TMPSSIZE];    // Char for the reason (might be too long)
  char cmdname[TMPSSIZE];   // Char for the command name (^^)
  char nickname[NICKLEN + 1]; // Char for the user's nickname
  char tmpstr[TMPSSIZE];    // Char for several purposes
  char target[10];      // String'y'fied numeric
  // Checking whether the user does have the right
  // auth level to be able to use this command.
  if (!checkauthlevel(unum,900)) { return; }
  // Simple error checking
  res = sscanf(tail, "%s %s %[^\n]", cmdname, tmpstr, reason);
  mystrncpy(nickname,tmpstr,NICKLEN);
  toLowerCase(nickname);
  if (res < 1) {
    putlog("!!! No commandname in opercommands.c !!!");
    return;
  } else if (res < 2) {
    sprintf(tmpstr, "Syntax: %s nick [reason]", cmdname);
    msgtouser(unum, tmpstr);
    return;
  }
  if (res == 2) {
    // No reason given
    udata = getudptr(unum);
    if (udata==NULL) { return; /* This should not happen! */ }
    sprintf(reason, "Requested by %s", udata->nick);
  }
  // Find the user *we* want to kill
  numeric = nicktonum(nickname);
  if (numeric < 0) {
    msgtouser(unum, "I don't know that user!");
    return;
  }
  longtotoken(numeric, target, 5);
  // Work
  fprintf(sockout, "%s D %s :%s %s\r\n", servernumeric, target, myname, reason);
  fflush(sockout);
  deluserfromallchans(numeric);
  delautheduser(numeric);
  if (!killuser(numeric)) {
    fakeuserkill(numeric);
    putlog("!!! KILL failed on dokillcmd: %s !!!", lastline);
    msgtouser(unum, "KILL was sent, but User could not be removed from my internal data. This is NOT good.");
  }
  msgtouser(unum, "KILL sent.");
}

void dospewchancmd(long unum, char *tail) {
  channel *tchan;       // Struct for channel data
  chanuser *cuser;      // Struct for one user of a channel
  userdata *udata;      // Struct for User Data as of User DB
  int res;          // Int for results
  int i, y;         // Int for loops
  int ccount;         // Int for couting the amount of channels found
  int gotchanserv;      // Int for flag usage
  int morechanservs;
  char cmdname[TMPSSIZE];   // Char for the command name (^^)
  char mask[TMPSSIZE];    // Char for the user's nickname
  char tmpstr[TMPSSIZE];    // Char for several purposes
  char tmpstr2[TMPSSIZE];   // Char for several purposes
  char chanservs[TMPSSIZE]; // Char for the namelist of all chanservs
  // Checking whether the user does have the right
  // auth level to be able to use this command.
  if (!checkauthlevel(unum,900)) { return; }
  // Simple error checking
  res = sscanf(tail, "%s %s", cmdname, mask);
  if (res < 1) {
    putlog("!!! No commandname in opercommands.c !!!");
    return;
  } else if (res < 2) {
    sprintf(tmpstr, "Syntax: %s mask", cmdname);
    msgtouser(unum, tmpstr);
    return;
  }
  toLowerCase(mask);
  ccount = 0;
  for (i = 0; i < SIZEOFCL; i++) {
    tchan = cls[i];
    while (tchan != NULL) {
      if (match2strings(mask, tchan->name)) {
        gotchanserv = 0; morechanservs=0;
        chanservs[0] = '\0';
        cuser = (chanuser *)(tchan->users.content);
        for (y=0; y < tchan->users.cursi; y++) {
            udata = getudptr(cuser[y].numeric);
            if (udata != NULL) {
            if (ischarinstr('k', udata->umode)) {
              // Channelservice on channel
              gotchanserv++;
              if (strlen(chanservs) < 300) {
                if (strlen(chanservs) > 0) { strcat(chanservs, ", "); }
                strcat(chanservs, udata->nick);
              } else {
                morechanservs++;
              }
            }
          }
        }
        if (morechanservs > 0) {
          // 300 letters exceeded, adding ...
          sprintf(tmpstr2,", and %d more",morechanservs);
          strcat(chanservs, tmpstr2);
        }
        sprintf(tmpstr2, "%d", gotchanserv);
        sprintf(tmpstr, "%-30s %5ld %-8s%s%s%-11s%s%s%s",
             tchan->name,
             tchan->users.cursi,
             (tchan->users.cursi > 1) ? "users" : "user",
             (gotchanserv == 0) ? "" : "- found ",
             (gotchanserv == 0) ? "" : tmpstr2,
             (gotchanserv == 0) ? "" : ((gotchanserv > 1) ? " chanservs" : " chanserv"),
             (gotchanserv == 0) ? "" : "(",
             (gotchanserv == 0) ? "" : chanservs,
             (gotchanserv == 0) ? "" : ")");
        if (ccount<SHOWTRUNC) { msgtouser(unum, tmpstr); }
        ccount++;
      }
      tchan = (void *) tchan->next;
    }
  }
  if (ccount >= SHOWTRUNC) {
    sprintf(tmpstr, "Found %i channels matching pattern %s. (only first %i shown)", ccount, mask, SHOWTRUNC);
  } else {
    sprintf(tmpstr, "Found %i channels matching pattern %s.", ccount, mask);
  }
  msgtouser(unum, tmpstr);
}

void dorawnumericcmd(long unum, char *tail) {
  // Define :]
  int res;          // Int for results
  long numeric;       // Long for numeric
  char token[10];       // Char for tokenized numeric
  char cmdname[TMPSSIZE];   // Char for the command name (^^)
  char nick[NICKLEN + 1];   // Char for the user's nickname
  char tmpstr[TMPSSIZE];    // Char for several purposes
  // Checking whether the user does have the right
  // auth level to be able to use this command.
  if (!checkauthlevel(unum, 999)) { return; }
  // Simple error checking
  res = sscanf(tail, "%s %s", cmdname, tmpstr);
  mystrncpy(nick,tmpstr,NICKLEN);
  toLowerCase(nick);
  if (res < 1) {
    putlog("!!! No commandname in opercommands.c !!!");
    return;
  } else if (res < 2) {
    sprintf(tmpstr, "Syntax: %s nick", cmdname);
    msgtouser(unum, tmpstr);
    return;
  }
  // Get the user's numeric and tokenize it
  numeric = nicktonum(nick);
  if (numeric < 0) {
    msgtouser(unum, "I don't know that user!");
    return;
  }
  longtotoken(numeric, token, 5);
  sprintf(tmpstr, "Data for %s: Numeric: %ld / Tokenized numeric: %s", nick, numeric, token);
  msgtouser(unum, tmpstr);
}

void dorawcmd(long unum, char *tail) {
  int res;          // Int for results
  char cmdname[TMPSSIZE];   // Char for the command name
  char raw[TMPSSIZE];     // Char for the user's nickname
  char tmpstr[TMPSSIZE];    // Char for several purposes
  char mode[TMPSSIZE];    // Char for several purposes
  // Checking whether the user does have the right
  // auth level to be able to use this command.
  if (!checkauthlevel(unum, 999)) { return; }
  // Simple error checking
  res = sscanf(tail, "%s %s %[^\n]", cmdname, mode, raw);
  if (res < 1) {
    putlog("!!! No commandname in opercommands.c !!!");
    return;
  } else if (res < 3) {
    sprintf(tmpstr, "Syntax: %s mode commands", cmdname);
    msgtouser(unum, tmpstr);
    msgtouser(unum, "where mode is either 'server' or 'client',");
    return;
  }
  toLowerCase(mode);
  if (strcmp(mode, "server") == 0) {
    strcpy(tmpstr, servernumeric);
  } else if (strcmp(mode, "client") == 0) {
    sprintf(tmpstr, "%sAAB", servernumeric);
  } else {
    msgtouser(unum, "There are two modes (first argument), server and client, use one of them");
    return;
  }
  /*
   We DO NOT perform ANY error checks here. If the user does
   send a BAD commandline (really bad I mean) servers might
   crash etc.
   Use this with CARE!
  */
  raw[450]='\0'; // At least truncate the command
  fprintf(sockout, "%s %s\r\n", tmpstr, raw);
  fflush(sockout);
  msgtouser(unum, "Raw done.");
}

void domkpasswdcmd (long unum, char *tail) {
  int res;          // Int for results
  char cmdname[TMPSSIZE];   // Char for the command name
  char plain[TMPSSIZE];   // Char for the plaintext password
  char tmpstr[TMPSSIZE];    // Char for several purposes
  char tmps2[TMPSSIZE];
  if (!checkauthlevel(unum, 900)) { return; }
  // Simple error checking
  res = sscanf(tail, "%s %s", cmdname, plain);
  if (res < 1) {
    putlog("!!! No commandname in opercommands.c !!!");
    return;
  } else if (res < 2) {
    sprintf(tmpstr, "Syntax: %s password", cmdname);
    msgtouser(unum, tmpstr);
    return;
  }
  encryptpwd(tmps2,plain);
  sprintf(tmpstr, "Plaintext: %s => Crypted: %s", plain, tmps2);
  msgtouser(unum, tmpstr);
}

void opercommands_init() {
  setmoduledesc(MODNAM, "Mod containing some oper commands originally offered by Q.");
  registercommand(MODNAM, "obroadcast", doobroadcastcmd, 1,
          "obroadcast text\tUsed to send a message to all oper'ed users");
  registercommand(MODNAM, "mbroadcast", dombroadcastcmd, 1,
          "mbroadcast text\tUsed to send a message to all users matching a certain mask");
  registercommand(MODNAM, "kill", dokillcmd, 1,
          "kill nick [reason]\tCommand to kill a certain user");
  registercommand(MODNAM, "spewchan", dospewchancmd, 1,
          "spewchan mask \tCommand to search channels by pattern");
  registercommand(MODNAM, "raw", dorawcmd, 1,
          "raw mode commands \tCommand to send raw server messages");
  registercommand(MODNAM, "rawnumeric", dorawnumericcmd, 1,
          "rawnumeric nick \tCommand to get the tokenized numeric of a user");
  registercommand(MODNAM, "mkpasswd", domkpasswdcmd, 1,
          "mkpasswd password \tCommand to encrypt a plaintext password");
}

void opercommands_cleanup() {
  deregistercommand("obroadcast");
  deregistercommand("mbroadcast");
  deregistercommand("kill");
  deregistercommand("spewchan");
  deregistercommand("raw");
  deregistercommand("rawnumeric");
  deregistercommand("mkpasswd");
}
