/*
Operservice 2 - trustcommands.c
Loadable module
(C) Michael Meier 2002 - released under GPL
-----------------------------------------------------------------------------
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
------------------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <regex.h>
#include "globals.h"
#ifndef WE_R_ON_BSD
#include <malloc.h>
#endif

/* Because the name of the mod is used in several places, we define it here so
   we can easily change it. It should be all lowercase. */
#define MODNAM "trustcommands"
/* Note: This .c-file will have to be named MODNAM.c! */

int trustlistsorthelper(const void *a, const void *b) {
  unsigned long *c, *d;
  c=(unsigned long *)a;
  d=(unsigned long *)b;
  return (*c - *d);
}

void displaytrustgroupwithid(long unum, unsigned long showid) {
  long i; trustedhost *th; trustedgroup *tg; char tmps2[TMPSSIZE], tmps3[TMPSSIZE];
  tg=findtrustgroupbyID(showid);
  if (tg==NULL) { return; }
  sprintf(tmps2,"Trustgroup '%s' (ID #%lu) added by %s - currently online: %d",tg->name,tg->id,tg->creator,tg->currentlyon);
  msgtouser(unum,tmps2);
  longtoduration(tmps3,tg->expires - getnettime());
  sprintf(tmps2,"Settings: Enforcing ident: %s - trusted for %d - %d per user - expires in %s",
    (tg->enforceident == 1) ? "yes" : "no ", tg->trustedfor, tg->maxperident, tmps3);
  msgtouser(unum,tmps2);
  sprintf(tmps2,"Contact: %s - Comment: %s",tg->contact,tg->comment);
  msgtouser(unum,tmps2);
  tmps2[0]=0;
  for (i=0;i<SIZEOFTL;i++) {
    th=trustedhosts[i];
    while (th!=NULL) {
      if (th->id==showid) {
        if (tmps2[0]==0) {
          strcpy(tmps2,th->hostname);
        } else {
          sprintf(tmps3,"|-%s",tmps2);
          msgtouser(unum,tmps3);
          strcpy(tmps2,th->hostname);
        }
      }
      th=(void *)th->next;
    }
  }
  if (tmps2[0]==0) {
    msgtouser(unum,"`- no hosts in that group");
  } else {
    sprintf(tmps3,"`-%s",tmps2); msgtouser(unum,tmps3);
  }
}

void dotrustgroupmodify(long unum, char *tail) {
  int res, newv, previous;
  char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE]; trustedgroup *tg;
  if (!checkauthlevel(unum, 900)) { return; }
  res = sscanf(tail, "%s %s %s", tmps2, tmps3, tmps4);
  if (res != 3) {
    msgtouser(unum, "Syntax: trustgroupmodify <groupname OR #groupid> [+|-|=]number");
    msgtouser(unum, "        +20 = add 20, -15 = subtract 15, =35 set it to 35");
    return;
  }
  toLowerCase(tmps3);
  if (tmps3[0] == '#') {
    tg = findtrustgroupbyID(strtol(&tmps3[1], NULL, 10));
  } else {
    tg = findtrustgroupbyname(tmps3);
  }
  if (tg == NULL) {
    sprintf(tmps2, "A trustgroup with that %s does not exist.", (tmps3[0] == '#') ? "ID" : "name");
    msgtouser(unum, tmps2);
    return;
  }
  if (tg->id == 0) {
    msgtouser(unum, "Internal error: Trustgroup has ID 0");
    return;
  }
  previous = tg->trustedfor;
  if (tmps4[0] == '=') {
    newv = strtol(tmps4 + 1, NULL, 10);
    if (newv > previous) {
      tmps4[0] = '+';
    } else {
      tmps4[0] = '-';
    }
  } else {
    newv = previous + strtol(tmps4, NULL, 10);
  }
  if (newv <= 0) {
    msgtouser(unum, "That would make it less or equal to 0 (zero).");
    return;
  }
  tg->trustedfor = newv;
  sprintf(tmps2, "Maxclones %s from %d to %d.", (tmps4[0] == '-') ? "lowered" : "raised", previous, newv);
  msgtouser(unum, tmps2);
}

void dotrustgroupdel(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE];
  trustedgroup *tg;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s %s",tmps2,tmps3,tmps4);
  if (res!=2) {
    msgtouser(unum,"Syntax: trustgroupdel <groupname OR #groupid>");
    msgtouser(unum,"This deletes the group including all hosts so be careful with this!");
    return;
  }
  toLowerCase(tmps3);
  if (tmps3[0]=='#') {
    tg=findtrustgroupbyID(strtol(&tmps3[1],NULL,10));
  } else {
    tg=findtrustgroupbyname(tmps3);
  }
  if (tg==NULL) {
    sprintf(tmps2,"A trustgroup with that %s does not exist.",(tmps3[0]=='#') ? "ID" : "name");
    msgtouser(unum,tmps2); return;
  }
  if (tg->id==0) { msgtouser(unum,"Internal error: Trustgroup has ID 0"); return; }
  numtonick(unum,tmps3);
  sprintf(tmps2,"%s deleted Trustgroup %s (#%ld)",tmps3,tg->name,tg->id);
  sendtonoticemask(NM_TRUSTS,tmps2);
  sprintf(tmps2,"Trustgroup with %d hosts deleted.",destroytrustgroup(tg->id));
  msgtouser(unum,tmps2);
}

void dotrustadd(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE];
  trustedgroup *tg; trustedhost *th; long i; userdata *u; char tmph[HOSTLEN+1];
  trustdeny * td;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s %s %s",tmps2,tmps3,tmps4,tmps5);
  if (res!=3) {
    msgtouser(unum,"Syntax: trustadd <groupname OR #groupid> host"); return;
  }
  toLowerCase(tmps3);
  if (tmps3[0]=='#') {
    tg=findtrustgroupbyID(strtol(&tmps3[1],NULL,10));
  } else {
    tg=findtrustgroupbyname(tmps3);
  }
  if (tg==NULL) {
    sprintf(tmps2,"A trustgroup with that %s does not exist.",(tmps3[0]=='#') ? "ID" : "name");
    msgtouser(unum,tmps2); return;
  }
  if (tg->id==0) { msgtouser(unum,"Internal error: Trustgroup has ID 0"); return; }
  toLowerCase(tmps4);
  if (strlen(tmps4)>HOSTLEN) {
    msgtouser(unum,"Hostname too long"); return;
  }
  if (istrusted(tmps4)>0) {
    msgtouser(unum,"That host already belongs to some trustgroup"); return;
  }
  if (ischarinstr('?',tmps4) || ischarinstr('*',tmps4)) {
    msgtouser(unum,"You cannot use wildcards in the host!"); return;
  }
  td=gettrustdeny(tmps4);
  if (td!=NULL) {
    if (td->type==TRUSTDENY_DENY) {
      msgtouser(unum,"Trusting this host is DENIED!");
    } else { /* Warn only */
      msgtouser(unum,"WARNING: There is a trustdeny-warning for this host.");
    }
    newmsgtouser(unum,"The entry was added by %s, reason: %s",td->creator,td->reason);
    if (td->type==TRUSTDENY_DENY) { return; }
  }
  /* OK so we have a valid group and host - create new hostentry */
  th=(trustedhost *)malloc(sizeof(trustedhost));
  strcpy(th->hostname,tmps4);
  th->id=tg->id;
  th->next=(void *)trustedhosts[tlhash(tmps4)];
  trustedhosts[tlhash(tmps4)]=th;
  /* Remove host from untrusted hosts, and instead add it to the trustgroup! */
  for (i=0;i<SIZEOFUL;i++) {
    u=uls[i];
    while (u!=NULL) {
      strcpy(tmph,u->host); toLowerCase(tmph);
      if (strcmp(tmph,tmps4)==0) {
        delfromuntrustedhost(tmps4);
        addtotrustedhost(u->ident,tg->id);
      }
      u=(void *)u->next;
    }
  }
  numtonick(unum,tmps3);
  sprintf(tmps2,"%s added host %s to trustgroup %s (#%lu)",tmps3,tmps4,tg->name,tg->id);
  sendtonoticemask(NM_TRUSTS,tmps2);
  sprintf(tmps2,"Added host %s to trustgroup %s (#%lu)",tmps4,tg->name,tg->id);
  msgtouser(unum,tmps2);
}

void dotrustdel(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE];
  trustedgroup *tg; trustedhost *th, *t2;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s %s",tmps2,tmps3,tmps4);
  if (res!=2) {
    msgtouser(unum,"Syntax: trustdel host"); return;
  }
  toLowerCase(tmps3);
  th=trustedhosts[tlhash(tmps3)]; t2=NULL;
  while (th!=NULL) {
    if (strcmp(th->hostname,tmps3)==0) {
      long i; userdata *u; char tmph[HOSTLEN+1];
      tg=findtrustgroupbyID(th->id);
      if (t2==NULL) {
        trustedhosts[tlhash(tmps3)]=(void *)th->next;
      } else {
        t2->next=th->next;
      }
      free(th);
      /* Recount the host */
      for (i=0;i<SIZEOFUL;i++) {
        u=uls[i];
        while (u!=NULL) {
          strcpy(tmph,u->host); toLowerCase(tmph);
          if (strcmp(tmph,tmps3)==0) {
            int currenton;
            delfromtrustedhost(u->ident,tg->id);
            currenton=addtountrustedhost(tmph);
            if (currenton>=defaultclonewarn) { /* Clonelimit for that host exceeded */
              if (defaultclonewarn!=0) {
                sprintf(tmps2,"[%d] clones detected from (previously trusted host) %s",currenton,u->host);
                noticeallircops(tmps2);
                putlog("[clones] %d clones detected from (now) untrusted host %s",currenton,u->host);
              }
            }
          }
          u=(void *)u->next;
        }
      }
      numtonick(unum,tmps4);
      sprintf(tmps2,"%s removed trust for Host %s",tmps4,tmps3);
      sendtonoticemask(NM_TRUSTS,tmps2);
      sprintf(tmps2,"Host removed from Trustgroup %s (#%lu)",tg->name,tg->id);
      msgtouser(unum,tmps2);
      return;
    } else {
      t2=th;
      th=(void *)th->next;
    }
  }
  msgtouser(unum,"That host was not trusted.");
}

void dotrustgroupadd(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE];
  char tmps6[TMPSSIZE], tmps7[TMPSSIZE], tmps8[TMPSSIZE], tmps9[TMPSSIZE];
  long dur; trustedgroup *tg;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s %s %s %s %s %s %[^\n]",tmps2,tmps3,tmps4,tmps5,tmps6,tmps7,tmps8,tmps9);
  if ((res<7) || (res>8)) {
    msgtouser(unum,"Syntax: trustgroupadd name howmany howlong maxperident enforceident contact [comment]");
    msgtouser(unum,"where name is a unique name for that group");
    msgtouser(unum,"      howmany gives the number of clones allowed from that group");
    msgtouser(unum,"      howlong gives how long the trust will last (e.g. '1y')");
    msgtouser(unum,"      maxperident sets how many clones are allowed from a single user.");
    msgtouser(unum,"                  0 means unlimited");
    msgtouser(unum,"      enforceident if set to 1 will kill all clients without idents");
    msgtouser(unum,"                   in that group");
    msgtouser(unum,"      contact is one or more contact-emails, seperated by commas,");
    msgtouser(unum,"              WITH NO SPACES!!!");
    msgtouser(unum,"      comment is an optional comment");
    return;
  }
  if (res==7) { strcpy(tmps9,"[no comment]"); }
  if (strlen(tmps3)>TRUSTNAMELEN) {
    msgtouser(unum,"The name you gave is too long");
    return;
  }
  if (tmps3[0]=='#') {
    msgtouser(unum,"Trustgroupname must not start with a '#'");
    return;
  }
  if (strlen(tmps8)>TRUSTCONTACTLEN) {
    msgtouser(unum,"contact-email is too long"); return;
  }
  if (strlen(tmps9)>TRUSTCOMLEN) {
    msgtouser(unum,"comment is too long"); return;
  }
  dur=durationtolong(tmps5);
  if (dur<1) {
    msgtouser(unum,"Invalid duration given");
    return;
  }
  if (((tmps7[0]!='0') && (tmps7[0]!='1')) || (tmps7[1]!='\0')) {
    msgtouser(unum,"enforceident is a boolean setting, that means it can only be 0 or 1"); return;
  }
  toLowerCase(tmps3);
  tg=findtrustgroupbyname(tmps3);
  if (tg!=NULL) { /* This group exists, modify it */
    sprintf(tmps2,"Trustgroup %s already exists, replacing values.",tmps3);
    msgtouser(unum,tmps2);
    sprintf(tmps2,"Old settings were: trusted for %d users, max. clones per user %d, %senforcing ident",
            tg->trustedfor,tg->maxperident,(tg->enforceident==0) ? "not " : "");
    msgtouser(unum,tmps2);
    sprintf(tmps2,"Contact-eMail: %s",tg->contact);
    msgtouser(unum,tmps2);
    sprintf(tmps2,"Comment:       %s",tg->comment);
    msgtouser(unum,tmps2);
  } else { /* New group */
    unsigned long newID;
    tg=(trustedgroup *)malloc(sizeof(trustedgroup));
    if (tg==NULL) {
      printf("!!! Out of memory in dotrustgroupadd !!!\n"); exit(0);
    }
    newID=getfreetgid();
    tg->id=newID;
    tg->currentlyon=0;
    tg->next=(void *)trustedgroups[tgshash(newID)];
    trustedgroups[tgshash(newID)]=tg;
    array_init(&(tg->identcounts),sizeof(identcounter));
  }
  tg->enforceident=(tmps7[0]=='1');
  strcpy(tg->name,tmps3);
  strcpy(tg->contact,tmps8);
  strcpy(tg->comment,tmps9);
  tg->trustedfor=strtol(tmps4,NULL,10);
  tg->maxperident=strtol(tmps6,NULL,10);
  tg->expires=getnettime()+dur;
  numtonick(unum,tmps2);
  mystrncpy(tg->creator,tmps2,AUTHUSERLEN);
  sprintf(tmps2,"%s created/updated Trustgroup %s (#%ld) - %d / %d / %d",
    tg->creator,tg->name,tg->id,tg->trustedfor,tg->maxperident,tg->enforceident);
  sendtonoticemask(NM_TRUSTS,tmps2);
  sprintf(tmps2,"Trustgroup %s (#%ld) created/updated",tg->name,tg->id);
  msgtouser(unum,tmps2);
}

void dotrustlist(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE];
  array listofids; long i; trustedgroup *tg; trustedhost *th; long j;
  unsigned long lastshown, showid;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s %s",tmps2,tmps3,tmps4);
  if (res>2) {
    msgtouser(unum,"Syntax: trustlist [pattern]");
    return;
  }
  if (res==1) { strcpy(tmps3,"*"); }
  toLowerCase(tmps3);
  trustgroupexpire();
  array_init(&listofids,sizeof(unsigned long));
  for (i=0;i<SIZEOFIDMAP;i++) {
    tg=trustedgroups[i];
    while (tg!=NULL) {
      strcpy(tmps4,tg->name); toLowerCase(tmps4);
      if (match2strings(tmps3,tmps4)) {
        j=array_getfreeslot(&listofids);
        ((unsigned long *)listofids.content)[j]=tg->id;
      }
      tg=(void *)tg->next;
    }
  }
  for (i=0;i<SIZEOFTL;i++) {
    th=trustedhosts[i];
    while (th!=NULL) {
      strcpy(tmps4,th->hostname); toLowerCase(tmps4);
      if (match2strings(tmps3,tmps4)) {
        j=array_getfreeslot(&listofids);
        ((unsigned long *)listofids.content)[j]=th->id;
      }
      th=(void *)th->next;
    }
  }
  if (listofids.cursi==0) {
    msgtouser(unum,"No matches.");
    array_free(&listofids);
    return;
  }
  qsort(listofids.content, listofids.cursi, sizeof(unsigned long), trustlistsorthelper);
  lastshown=0;
  for (i=0;i<listofids.cursi;i++) {
    showid=((unsigned long *)(listofids.content))[i];
    if (showid!=lastshown) {
      lastshown=showid;
      displaytrustgroupwithid(unum,showid);
    }
  }
  array_free(&listofids);
}

void dotrustgline(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE];
  char tmps6[TMPSSIZE]; long r2; userdata *a;
  trustedgroup *tg; trustedhost *th; int i, j; long l; array listofths;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail, "%s %s %s %s %[^\n]",tmps2,tmps3,tmps4,tmps5,tmps6);
  if (res<4) {
    msgtouser(unum,"Syntax: trustgline <groupname OR #groupID> user duration [reason]");
    msgtouser(unum,"user may contain wildcards");
    return;
  }
  if (res==4) {
    numtonick(unum,tmps2);
    sprintf(tmps6,"requested by %s",tmps2);
  }
  tmps4[USERLEN+2]='\0';
  toLowerCase(tmps3); toLowerCase(tmps4);
  r2=durationtolong(tmps5);
  if (r2<=0) {
    sprintf(tmps2,"The duration you gave is invalid.");
    msgtouser(unum,tmps2);
    return;
  }
  if (tmps3[0]=='#') {
    tg=findtrustgroupbyID(strtol(&tmps3[1],NULL,10));
  } else {
    tg=findtrustgroupbyname(tmps3);
  }
  if (tg==NULL) {
    sprintf(tmps2,"A trustgroup with that %s does not exist.",(tmps3[0]=='#') ? "ID" : "name");
    msgtouser(unum,tmps2); return;
  }
  /* First, create a list of all hosts in that trustgroup... */
  array_init(&listofths,HOSTLEN+1);
  for (i=0;i<SIZEOFTL;i++) {
    th=trustedhosts[i];
    while (th!=NULL) {
      if (th->id==tg->id) { /* mmkay, that one belongs to our group */
        long j;
        j=array_getfreeslot(&listofths);
        strcpy(((char *)listofths.content)+(j*(HOSTLEN+1)),th->hostname);
        toLowerCase(((char *)listofths.content)+(j*(HOSTLEN+1)));
      }
      th=(void *)th->next;
    }
  }
  if (listofths.cursi==0) {
    msgtouser(unum,"There are no hosts in that trustgroup.");
    array_free(&listofths);
    return;
  }
  /* Now that we have the list: count number of users hit */
  l=0;
  for (j=0;j<SIZEOFUL;j++) {
    a=uls[j];
    while (a!=NULL) {
      char tmph[HOSTLEN+1], tmpu[USERLEN+1];
      strcpy(tmph,a->host); strcpy(tmpu,a->ident);
      toLowerCase(tmph); toLowerCase(tmpu);
      for (i=0;i<listofths.cursi;i++) {
        char *bla;
        bla=(char *)listofths.content;
        if (strcmp(tmph,bla+(i*(HOSTLEN+1)))==0) {
          if (match2strings(tmps4,tmpu)) {
            l++; break;
          }
        }
      }
      a=(void *)a->next;
    }
  }
  if (l>500) {
    sprintf(tmps2,"That gline would hit more than 500 (%ld) users. You probably mistyped something.",l);
    msgtouser(unum,tmps2);
    array_free(&listofths); return;
  }
  /* OK, safety checks done - now gline the fuckers */
  numtonick(unum,tmps5);
  for (i=0;i<listofths.cursi;i++) {
    char *bla;
    bla=(char *)listofths.content;
    sprintf(tmps2,"%s@%s",tmps4,bla+(i*(HOSTLEN+1)));
    addgline(tmps2,tmps6,tmps5,r2,1);
  }
  sprintf(tmps2,"Added GLINE for %s @ trustgroup %s (#%lu), caused %lu glines, hit %ld users",tmps4,tg->name,tg->id,
    listofths.cursi,l);
  msgtouser(unum,tmps2);
  sprintf(tmps2,"GLINE %s @ trustgroup %s (#%lu) by %s, caused %lu glines, hit %ld users",tmps4,
    tg->name,tg->id,tmps5,listofths.cursi,l);
  sendtonoticemask(NM_GLINE,tmps2);
  array_free(&listofths);
}

void dotrustungline(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE];
  trustedgroup *tg; trustedhost *th; int i; array listofths;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail, "%s %s %s",tmps2,tmps3,tmps4);
  if (res<3) {
    msgtouser(unum,"Syntax: trustungline <groupname OR #groupID> user");
    msgtouser(unum,"user may contain wildcards");
    return;
  }
  tmps4[USERLEN+2]='\0';
  toLowerCase(tmps3); toLowerCase(tmps4);
  if (tmps3[0]=='#') {
    tg=findtrustgroupbyID(strtol(&tmps3[1],NULL,10));
  } else {
    tg=findtrustgroupbyname(tmps3);
  }
  if (tg==NULL) {
    sprintf(tmps2,"A trustgroup with that %s does not exist.",(tmps3[0]=='#') ? "ID" : "name");
    msgtouser(unum,tmps2); return;
  }
  /* First, create a list of all hosts in that trustgroup... */
  array_init(&listofths,HOSTLEN+1);
  for (i=0;i<SIZEOFTL;i++) {
    th=trustedhosts[i];
    while (th!=NULL) {
      if (th->id==tg->id) { /* mmkay, that one belongs to our group */
        long j;
        j=array_getfreeslot(&listofths);
        strcpy(((char *)listofths.content)+(j*(HOSTLEN+1)),th->hostname);
      }
      th=(void *)th->next;
    }
  }
  if (listofths.cursi==0) {
    msgtouser(unum,"There are no hosts in that trustgroup.");
    array_free(&listofths);
    return;
  }
  /* starting ungline.. */
  numtonick(unum,tmps5);
  for (i=0;i<listofths.cursi;i++) {
    char *bla;
    bla=(char *)listofths.content;
    sprintf(tmps2,"%s@%s",tmps4,bla+(i*(HOSTLEN+1)));
    remgline(tmps2,1);
  }
  sprintf(tmps2,"Removed GLINE for %s @ trustgroup %s (#%lu)",tmps4,tg->name,tg->id);
  msgtouser(unum,tmps2);
  sprintf(tmps2,"GLINE %s @ trustgroup %s (#%lu) removed by %s",tmps4, tg->name,tg->id,tmps5);
  sendtonoticemask(NM_GLINE,tmps2);
  array_free(&listofths);
}

void dotrustspew(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE];
  trustedgroup *tg; trustedhost *th; int i; userdata *ud; long k, l;
  array listofths; int iamtired;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail, "%s %s %s",tmps2,tmps3,tmps4);
  if (res<2) {
    msgtouser(unum,"Syntax: trustspew <groupname OR #groupID> [user]");
    return;
  }
  if (res==2) { strcpy(tmps4,"*"); }
  tmps4[USERLEN+2]='\0';
  toLowerCase(tmps3); toLowerCase(tmps4);
  if (tmps3[0]=='#') {
    tg=findtrustgroupbyID(strtol(&tmps3[1],NULL,10));
  } else {
    tg=findtrustgroupbyname(tmps3);
  }
  if (tg==NULL) {
    sprintf(tmps2,"A trustgroup with that %s does not exist.",(tmps3[0]=='#') ? "ID" : "name");
    msgtouser(unum,tmps2); return;
  }
  /* First, create a list of all hosts in that trustgroup... */
  array_init(&listofths,HOSTLEN+1);
  for (i=0;i<SIZEOFTL;i++) {
    th=trustedhosts[i];
    while (th!=NULL) {
      if (th->id==tg->id) { /* mmkay, that one belongs to our group */
        long j;
        j=array_getfreeslot(&listofths);
        strcpy(((char *)listofths.content)+(j*(HOSTLEN+1)),th->hostname);
      }
      th=(void *)th->next;
    }
  }
  if (listofths.cursi==0) {
    msgtouser(unum,"There are no hosts in that trustgroup.");
    array_free(&listofths);
    return;
  }
  /* We have the list - scan through the whole userlist, and match against our
     hostlist, and if that matches, match again against the username */
  sprintf(tmps2,"Users in Trustgroup %s (#%lu) matching %s",tg->name,tg->id,tmps4);
  msgtouser(unum,tmps2);
  l=0;
  for (i=0;i<SIZEOFUL;i++) {
    ud=uls[i];
    while (ud!=NULL) {
      strcpy(tmps2,ud->host);
      toLowerCase(tmps2);
      for (k=0;k<listofths.cursi;k++) {
        if (strcmp(tmps2,((char *)listofths.content)+(k*(HOSTLEN+1)))==0) {
          /* Host matches, check username */
          strcpy(tmps5,ud->ident);
          toLowerCase(tmps5);
          if (match2strings(tmps4,tmps5)) {
            /* yay ident matches too - list that loser */
            channel **d; int chanmodes;
            l++; /* inc counter of number of users found */
            sprintf(tmps2,"%s!%s@%s (%s)",ud->nick,ud->ident,ud->host,ud->realname);
            msgtouser(unum,tmps2);
            if (ud->chans.cursi>0) {
              if (ud->chans.cursi==1) {
                strcpy(tmps2,"On channel:");
              } else {
                strcpy(tmps2,"On channels:");
              }
              for (iamtired=0;iamtired<(ud->chans.cursi);iamtired++) {
                d=(channel **)(ud->chans.content);
                if ((strlen(d[iamtired]->name)+strlen(tmps2))>400) {
                  strcat(tmps2," [...]"); break;
                } else {
                  chanmodes=getchanmode2(d[iamtired],ud->numeric);
                  strcat(tmps2," ");
                  if (chanmodes<0) {
                    strcat(tmps2,"?");
                  } else {
                    if (isflagset(chanmodes,um_o)) { strcat(tmps2,"@"); } else { strcat(tmps2," "); }
                  }
                  strcat(tmps2,d[iamtired]->name);
                }
              }
            } else {
              strcpy(tmps2,"Not on any (nonlocal-)channels");
            }
            msgtouser(unum,tmps2);
          }
        }
      }
      ud=(void *)ud->next;
    }
  }
  sprintf(tmps2,"--- End of list - %ld matches ---",l);
  msgtouser(unum,tmps2);
  array_free(&listofths);
}

void dotrustdenyadd(long unum, char *tail) {
  int res; long i; trustdeny *td; long expires;
  char tmps2[TMPSSIZE], tmps3[TMPSSIZE], tmps4[TMPSSIZE], tmps5[TMPSSIZE], tmps6[TMPSSIZE];
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s %s %s %[^\n]",tmps2,tmps3,tmps4,tmps5,tmps6);
  if (res<5) {
    msgtouser(unum,"Syntax: trustdenyadd warn|deny hostmask duration reason");
    return;
  }
  toLowerCase(tmps3); toLowerCase(tmps4);
  if (strcmp(tmps3,"warn")==0) { res=TRUSTDENY_WARN; }
  if (strcmp(tmps3,"deny")==0) { res=TRUSTDENY_DENY; }
  if ((res!=TRUSTDENY_WARN) && (res!=TRUSTDENY_DENY)) {
    msgtouser(unum,"You can only warn or deny.");
    return;
  }
  expires=durationtolong(tmps5);
  if (expires<1) {
    msgtouser(unum,"Invalid duration.");
    return;
  }
  expires+=getnettime();
  td=(trustdeny *)deniedtrusts.content;
  for (i=0;i<deniedtrusts.cursi;i++) {
    if (strcmp(tmps4,td[i].host)==0) {
      longtoduration(tmps2,getnettime()-td[i].created);
      longtoduration(tmps3,td[i].expires-getnettime());
      msgtouser(unum,"a trustdeny for that hostmask already exists - replacing values");
      newmsgtouser(unum,"Old one was created by %s %s ago, expiring in %s and mode %s",
        td[i].creator,tmps2,tmps3,(td[i].type==TRUSTDENY_WARN) ? "WARN" : "DENY");
      getauthedas(tmps2,unum);
      strcpy(td[i].creator,tmps2);
      strcpy(td[i].reason,tmps6);
      td[i].expires=expires;
      td[i].created=getnettime();
      td[i].type=res;
      msgtouser(unum,"Done.");
      return;
    }
  }
  /* Not existing yet - allocate new entry */
  i=array_getfreeslot(&deniedtrusts);
  td=(trustdeny *)deniedtrusts.content;
  getauthedas(tmps2,unum);
  strcpy(td[i].creator,tmps2);
  strcpy(td[i].reason,tmps6);
  td[i].expires=expires;
  td[i].created=getnettime();
  strcpy(td[i].host,tmps4);
  td[i].type=res;
  msgtouser(unum,"Done.");
}

void dotrustdenylist(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE]; long i, j; trustdeny *td;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s",tmps2,tmps3);
  if (res<2) { strcpy(tmps3,"*"); }
  toLowerCase(tmps3);
  trustdenyexpire();
  td=(trustdeny *)deniedtrusts.content;
  j=0;
  for (i=0;i<deniedtrusts.cursi;i++) {
    if (match2strings(tmps3,td[i].host)) {
      if (j==0) {
        newmsgtouser(unum,"%-30s %-15s %-17s %-1s %s","Host","by","expires in","t","Reason");
      }
      longtoduration(tmps2,td[i].expires-getnettime());
      newmsgtouser(unum,"%-30s %-15s %-17s %s %s",td[i].host,td[i].creator,
        tmps2,(td[i].type==TRUSTDENY_DENY) ? "D" : "W", td[i].reason);
      j++;
    }
  }
  if (j==0) {
    msgtouser(unum,"--- No matches ---");
  } else {
    newmsgtouser(unum,"--- End of list - %ld matches ---",j);
  }
}

void dotrustdenydel(long unum, char *tail) {
  int res; char tmps2[TMPSSIZE], tmps3[TMPSSIZE]; long i; trustdeny * td;
  if (!checkauthlevel(unum,900)) { return; }
  res=sscanf(tail,"%s %s",tmps2,tmps3);
  if (res<2) {
    msgtouser(unum,"Syntax: trustdenydel hostmask");
    return;
  }
  toLowerCase(tmps3);
  td=(trustdeny *)deniedtrusts.content;
  for (i=0;i<deniedtrusts.cursi;i++) {
    if (strcmp(tmps3,td[i].host)==0) {
      array_delslot(&deniedtrusts,i);
      msgtouser(unum,"trustdeny removed.");
      return;
    }
  }
  msgtouser(unum,"No such trustdeny");
}

void trustcommands_init(void) {
  setmoduledesc(MODNAM,"This mod provides the trust*-commands");
  registercommand(MODNAM,"trustgroupadd", dotrustgroupadd, 1,
    "trustgroupadd\tAdds a new or modifies an existing trustgroup");
  registercommand(MODNAM,"trustgroupmodify", dotrustgroupmodify, 1,
    "trustgroupmodify name n\tChanges trustlimit for a trustgroup");
  registercommand(MODNAM,"trustgroupdel", dotrustgroupdel, 1,
    "trustgroupdel name|#gid\tDeletes a trustgroup (be careful with that!)");
  registercommand(MODNAM,"trustadd", dotrustadd, 1,
    "trustadd name|#gid host\tAdds a host to a trustgroup");
  registercommand(MODNAM,"trustdel", dotrustdel, 1,
    "trustdel host\tRemoves a host from a trustgroup");
  registercommand(MODNAM,"trustlist", dotrustlist, 1,
    "trustlist [pattern]\tLists matching or all trustgroups/hosts");
  registercommand(MODNAM,"trustgline", dotrustgline, 1,
    "trustgline ...\tGLINEs a user on all hosts of a trustgroup");
  registercommand(MODNAM,"trustungline", dotrustungline, 1,
    "trustungline ...\tRemoves GLINEs of a user on all hosts of a trustgroup");
  registercommand(MODNAM,"trustspew", dotrustspew, 1,
    "trustspew group [user]\tLists users on a trustgroup");
  registercommand(MODNAM,"trustdenyadd", dotrustdenyadd, 1,
    "trustdenyadd ...\tAdds a trustdeny to a hostmask");
  registercommand(MODNAM,"trustdenylist", dotrustdenylist, 1,
    "trustdenylist [pattern]\tLists existing trustdenys");
  registercommand(MODNAM,"trustdenydel", dotrustdenydel, 1,
    "trustdenydel hostmask\tRemoves a trustdeny");
}

/* This is called for cleanup (before the module gets unloaded)
   Again the name of the function has to be MODNAM_cleanup */
void trustcommands_cleanup(void) {
  deregistercommand("trustgroupadd");
  deregistercommand("trustgroupmodify");
  deregistercommand("trustgroupdel");
  deregistercommand("trustadd");
  deregistercommand("trustdel");
  deregistercommand("trustlist");
  deregistercommand("trustgline");
  deregistercommand("trustungline");
  deregistercommand("trustspew");
  deregistercommand("trustdenyadd");
  deregistercommand("trustdenylist");
  deregistercommand("trustdenydel");
}
