/*
Operservice 2
(C) Michael Meier 2000-2001 - released under GPL
-----------------------------------------------------------------------------
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
------------------------------------------------------------------------------
*/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <errno.h>
#include <sys/types.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "globals.h"

/* Subnetlist-Management-Functions. */

int snlcreate(subnetlist *snl, long mysize) {
  snl->maxsize=mysize;
  snl->cursize=0;
  snl->s=(subnetdata *) malloc(mysize*sizeof(subnetdata));
  if (snl->s==NULL) { return -1; }
  return 0;
}

void snlfree(subnetlist *snl) {
  free(snl->s);
  snl->maxsize=0;
  snl->cursize=0;
}

int snlcomp(subnetdata *a, subnetdata *b) {
  if (a->n==b->n) { return 0; }
  if (a->n>b->n) { return -1; } else { return 1; }
}

void snlsort(subnetlist *snl) {
  if (snl->cursize==0) { return; }
  qsort(snl->s,snl->cursize,sizeof(subnetdata),(void *)snlcomp);
}

unsigned int snladd(subnetlist *snl, unsigned long IP) {
  int i; unsigned int a, b, c, d; subnetdata *s, *p;
  if ((snl->cursize+1)>=snl->maxsize) {
    // Resize the list
    snl->maxsize+=100;
    snl->s=realloc((void *) snl->s,snl->maxsize*sizeof(subnetdata));
    if (snl->s==NULL) {
      putlog("!!! Subnetlist: Memory Reallocation failed. !!!");
      exit(-1);
    }
  }
  s=snl->s;
  splitip(IP,&a,&b,&c,&d);
  for (i=0;i<snl->cursize;i++) {
    p=s+i;
    if ((p->a==a) && (p->b==b) && (p->c==c)) {
      p->n++;
      return (p->n);
    }
  }
  // That subnet is not yet in our list. Add it.
  p=s+snl->cursize;
  p->a=a;
  p->b=b;
  p->c=c;
  p->n=1;
  snl->cursize++;
  return 1;
}

void snldel(subnetlist *snl, unsigned long IP) {
  unsigned int a, b, c, d; subnetdata *s, *p; int i;
  s=snl->s;
  splitip(IP,&a,&b,&c,&d);
  for (i=0;i<snl->cursize;i++) {
    p=&s[i];
    if ((p->a==a) && (p->b==b) && (p->c==c)) {
      p->n--;
      if (p->n<=0) {
        // This subnet now has no Users left, so we have to kill it from our list
        snl->cursize--;
        s[i]=s[snl->cursize];
      }
      return;
    }
  }
}

/* End of Subnetlist-Management-Functions. */
